@@ -393,7 +393,7 @@ CAMLexport value caml_cairo_path_fold(value vpath, value fn, value va)
   for(i = 0; i < path->num_data; i += path->data[i].header.length) {
     data = &path->data[i];
     PATH_DATA_ASSIGN(vdata, data);
-    vdata = caml_callback2(fn, vacc, vdata);
+    vacc = caml_callback2(fn, vacc, vdata);
   }
   CAMLreturn(vacc);
 }
@@ -1006,7 +1006,7 @@ CAMLexport value caml_cairo_scaled_font_text_to_glyphs
 
   status = cairo_scaled_font_text_to_glyphs
     (SCALED_FONT_VAL(vsf), Double_val(vx), Double_val(vy),
-     String_val(vutf8), string_length(vutf8),
+     String_val(vutf8), caml_string_length(vutf8),
      &glyphs, &num_glyphs,  &clusters, &num_clusters,  &cluster_flags);
   caml_cairo_raise_Error(status);
 
@@ -1269,7 +1269,7 @@ CAMLexport value caml_cairo_show_text_glyphs
 
   ARRAY_GLYPH_VAL(glyphs, p, vglyphs, num_glyphs);
   ARRAY_CLUSTER_VAL(clusters, q, vglyphs, num_glyphs);
-  cairo_show_text_glyphs(cr, String_val(vutf8), string_length(vutf8),
+  cairo_show_text_glyphs(cr, String_val(vutf8), caml_string_length(vutf8),
                          glyphs, num_glyphs, clusters, num_clusters,
                          /* FIXME: is it a binary | ? */
                          CLUSTER_FLAGS_VAL(vcluster_flags));
@@ -1564,6 +1564,7 @@ static cairo_status_t caml_cairo_image_bigarray_attach_proxy
     return(CAIRO_STATUS_SUCCESS);
   if (b->proxy != NULL) {
     /* If b is already a proxy, increment refcount. */
+    /* FIXME, use caml_atomic_refcount_incr if available. */
     ++ b->proxy->refcount;
   }
   else {
@@ -1623,14 +1624,20 @@ SURFACE_CREATE_DATA(data32)
       cairo_surface_get_user_data(SURFACE_VAL(vsurf), &image_bigarray_key); \
                                                                         \
     if (data == NULL)                                                   \
-      invalid_argument("Cairo.Image.get_data: not an image surface.");  \
-    if (proxy == NULL)                                                  \
-      invalid_argument("Cairo.Image.get_data: not created from a bigarray"); \
-    vb = caml_ba_alloc(CAML_BA_##type | CAML_BA_C_LAYOUT | CAML_BA_MANAGED, \
-                       num_dims, data, dim);                            \
-    /* Attach the proxy of the surface to the bigarray */               \
-    ++ proxy->refcount;                                                 \
-    (Caml_ba_array_val(vb))->proxy = proxy;                             \
+      caml_invalid_argument("Cairo.Image.get_data: not an image surface.");  \
+    if (proxy == NULL) {                                                \
+      /* We assume the payload is externally managed */                \
+      vb = caml_ba_alloc(CAML_BA_##type | CAML_BA_C_LAYOUT              \
+                         | CAML_BA_EXTERNAL,                            \
+                         num_dims, data, dim);                          \
+    } else {                                                            \
+      vb = caml_ba_alloc(CAML_BA_##type | CAML_BA_C_LAYOUT              \
+                         | CAML_BA_MANAGED,                             \
+                         num_dims, data, dim);                          \
+      /* Attach the proxy of the surface to the bigarray */             \
+      ++ proxy->refcount;                                               \
+      (Caml_ba_array_val(vb))->proxy = proxy;                           \
+    }                                                                   \
     CAMLreturn(vb);                                                     \
   }
 
